!c
!c module of mesh data and other functions
!c
module mesh_data

#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscdmda.h>
#include <petsc/finclude/petscksp.h>

  use petscsys
  use petscdmda
  use petscksp

  implicit none

  !c point data type
  type :: point
    real*8 :: x
    real*8 :: y
    real*8 :: z
  end type point

  type userctx
    !> DMDA
    DM :: dmx
    !> degree of freedom
    PetscInt :: dof
    !> stencil width
    PetscInt :: swidth
    !> dimension of the distributed array
    PetscInt :: dim
    !> global Vector ranger, 1-based  
    PetscInt :: range_start, range_end
  end type userctx   

  !c mesh variables
  integer :: ndim
  integer :: num_nodes
  integer :: num_cells
  integer :: num_nodes_per_cell
  integer, allocatable :: cells(:,:)
  type(point), allocatable :: nodes(:)

  contains

  !c
  !c read 2d triangular mesh from vtk file in ascii format
  !c
  subroutine read_mesh_vtk(rank, strfile)
  
    implicit none  
  
    integer, intent(in) :: rank
    character(*), intent(in) :: strfile
  
    !c local variables
    integer :: i, j, itype, ivtk, idummy, itemp, ntemp
    character(256) :: strbuffer
  
    integer, parameter :: cell_type_tri = 5
  
 
    !c initialization for 2d triangular mesh
    ndim = 0
    num_nodes_per_cell = 0
    num_nodes = 0
    num_cells = 0  
  
    !c open file
    ivtk = 100
    open(ivtk,file=trim(strfile),status='old',form='formatted',    &
         access='sequential',err=1)
  
    !c check number of triangular cells first
    do while(.true.)
      !c begin reading dataset, keywords: points, cells, cell_types followed by the number of entities
      read(ivtk,'(a)',err=1,end=1) strbuffer
      call makelowercase(strbuffer)      
      if(index(strbuffer,'cell_types') == 1) then
        i = index(strbuffer,' ')
        if(i > 0) then
          read(strbuffer(i:),*) ntemp
        else
          if(rank == 0) then
            write(*,*) "error in reading number of cell_types"
          end if
          goto 1
        end if
  
        do i = 1, ntemp
          read(ivtk,'(i2)',err=1,end=1) itype
          if (itype == cell_type_tri) then
            num_cells = num_cells + 1
          end if
        end do
        exit
      end if
    end do
  
    if (num_cells < 1) then
      if (rank == 0) then
        write(*,*) "number of cells is less than 1, return"
      end if
      return
    else
      ndim = 2
      num_nodes_per_cell = 3
    end if
  
    rewind(ivtk)
  
    !c read the version of mesh data file
    read(ivtk,'(a)',err=1,end=1) strbuffer
    call makelowercase(strbuffer)
  
    if(index(strbuffer,'# vtk datafile version') /= 1) then
      if(rank == 0) then
        write(*,*) "error reading in mesh data version"
      end if
      goto 1
    end if
  
    !c read the header of mesh data file
    read(ivtk,'(a)',err=1,end=1) strbuffer
    call makelowercase(strbuffer)
  
    !c read the file format, currently only support ascii
    read(ivtk,'(a)',err=1,end=1) strbuffer
    call makelowercase(strbuffer)
  
    if(index(strbuffer,'ascii') /= 1) then
      if(rank == 0) then
        write(*,*) "error: vtk file format should be ascii"
      end if
      goto 1
    end if

    if (rank == 0) then
      write(*,*) "start reading cell-node dataset"
    end if
  
    !c read the dataset 
    read(ivtk,'(a)',err=1,end=1) strbuffer
    call makelowercase(strbuffer)

    do while(.true.)
  
      !c begin reading dataset, keywords: points, cells, cell_types followed by the number of entities
      read(ivtk,'(a)',err=1,end=1) strbuffer
      call makelowercase(strbuffer)
  
      if(index(strbuffer,'points') == 1) then
        i = index(strbuffer,' ')
        if(i > 0) then
          read(strbuffer(i:),*) num_nodes      
        else
          if(rank == 0) then
            write(*,*) "error in reading number of points"
          end if
          goto 1
        end if
  
        !c allocate nodes dataset
        allocate(nodes(num_nodes))
        do i = 1, num_nodes
          read(ivtk,*) nodes(i)%x,nodes(i)%y,nodes(i)%z
        end do

        if (rank == 0) then
          write(*,*) "finish reading vertex data"
        end if
  
      else if(index(strbuffer,'cells') == 1) then
        i = index(strbuffer,' ')
        if(i > 0) then
          read(strbuffer(i:),*) ntemp, idummy
        else
          if(rank == 0) then
            write(*,*) "error in reading number of cells"
          end if
          goto 1
        end if
  
        !c allocate cells dataset
        allocate(cells(num_nodes_per_cell, num_cells))
        i = 0
        do j = 1, ntemp
          read(ivtk,'(a)',err=1,end=1) strbuffer
          !c only read selected cell type
          read(strbuffer,*) itemp
          if (itemp == num_nodes_per_cell) then
            i = i + 1
            read(strbuffer,*) idummy, cells(1:num_nodes_per_cell,i)
  
            !c convert 0-based to 1-based
            cells(1:num_nodes_per_cell,i) = cells(1:num_nodes_per_cell,i) + 1
          end if
        end do

        if (rank == 0) then
          write(*,*) "finish reading cell data"
        end if
      end if

      !c exit data reading when both nodes and cells are assigned.
      if (allocated(nodes) .and. allocated(cells)) then
        exit
      end if

    end do
  
    !c close file after reading
    close(ivtk)
  
    return
  
1   continue
    if(rank == 0) then
      write(*,*) "Error detected in file reading, return"
    end if
  
  end subroutine read_mesh_vtk
  
  !c
  !c Convert upper case to lower case
  !c
  subroutine makelowercase(string)
  
    implicit none
  
    character(*), intent(INOUT) :: string
  
    integer :: i,j
  
    do i = 1,len_trim(string)
      j = ichar(string(i:i))
      if (j >= 65 .and. j <= 90) then
        j = j + 32
        string(i:i) = char(j)
      end if
    end do
  
  end subroutine makelowercase

end module mesh_data
