!c ----------------------------------------------------------------------
!c example to read unstructured mesh from vtk file, distribute and check
!c global - natural ordering in DMPlex
!c ----------------------------------------------------------------------

!c
!c main program
!c
program driver_pc

#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscvec.h>
#include <petsc/finclude/petscdm.h>
#include <petsc/finclude/petscis.h>
#include <petsc/finclude/petscdmplex.h>
  use petscsys
  use petscvec
  use petscdm
  use petscis
  use petscdmplex

  use mesh_data

  implicit none

  !c system variables
  integer :: rank, nprcs
  PetscErrorCode :: ierr

  !c mesh file path
  character(8) :: strfile

  !c petsc dmplex related variables
  integer, allocatable :: dmplex_cells(:)
  real*8, allocatable :: dmplex_verts(:)

  type(userctx) :: dmda_flow
  DM :: pdm, cdm
  Vec :: gc, vec_global, vec_natural
  PetscSF :: migrationsf
  PetscScalar :: scalar
  IS :: globalvtx
  PetscInt :: ipoint, istart, iend, idof, dof, off
  PetscInt, pointer :: array(:)
  PetscScalar, pointer :: vecpointer(:)
  PetscScalar, pointer :: coords(:)

  PetscSection :: section
  PetscInt :: numFields, numBC
  PetscInt, target, dimension(1) ::  numComp
  PetscInt, target, dimension(4) ::  numDof
  PetscInt, target, dimension(1) ::  bcField

  PetscInt, pointer :: pNumComp(:)
  PetscInt, pointer :: pNumDof(:)
  PetscInt, pointer :: pBcField(:)
  IS, target, dimension(1) ::   bcCompIS
  IS, target, dimension(1) ::   bcPointIS
  IS, pointer :: pBcCompIS(:)
  IS, pointer :: pBcPointIS(:)

  !c other variables
  integer :: i, inode, icell, icnt

  !c initialize PETSc
  call PetscInitialize(Petsc_Null_Character,ierr)  
  CHKERRQ(ierr)
  call MPI_Comm_rank(Petsc_Comm_World,rank,ierr)
  CHKERRQ(ierr)
  call MPI_Comm_size(Petsc_Comm_World,nprcs,ierr)
  CHKERRQ(ierr)

  !c check initialization
  !write(*,"(2(a,1x,i4,1x))") "rank",rank,"of", nprcs

  !c read mesh from vtk file
  if (rank == 0) then
    write(*,*) "read 2d triangular mesh from vtk file"
  end if

  strfile = "mesh.vtk"
  call read_mesh_vtk(rank, strfile)

  !c check mesh data
  !c all the indices have been converged to 1-based.
  if (rank == 0) then
    do i = 1, num_nodes
      write(*,"(a,1x,i4,1x,a,3(1x,e12.5))") "node",i,"coord",nodes(i)%x,nodes(i)%y,nodes(i)%z
    end do

    do i = 1, num_cells
      write(*,"(a,1x,i4,1x,a,3(1x,i4))") "cell",i,"nodes",cells(1,i),cells(2,i),cells(3,i)
    end do
  end if

  !c create cell list for dmplex
  if (rank == 0) then
    allocate(dmplex_cells(num_cells*num_nodes_per_cell))
    allocate(dmplex_verts(num_nodes*ndim))

    !c set cell data
    do icell = 1, num_cells
      do inode = 1, num_nodes_per_cell
        icnt = (icell-1)*num_nodes_per_cell+inode
        dmplex_cells(icnt) = cells(inode,icell)-1        !convert back to 0-based
      end do
    end do

    !c set coordinates, for 2D XZ coordinate only
    do inode = 1, num_nodes
      icnt = (inode-1)*ndim                              !convert back to 0-based
      dmplex_verts(icnt + 1) = nodes(inode)%x
      dmplex_verts(icnt + 2) = nodes(inode)%z
    end do

    call DMPlexCreateFromCellList(Petsc_Comm_World,ndim,num_cells, &
                                  num_nodes,num_nodes_per_cell,    &
                                  Petsc_False,dmplex_cells,ndim,   &
                                  dmplex_verts,dmda_flow%dmx,ierr)
    CHKERRQ(ierr)

    deallocate(dmplex_cells)
    deallocate(dmplex_verts)

  else

    call DMPlexCreateFromCellList(Petsc_Comm_World,ndim,0,0,       &
                                  num_nodes_per_cell,              &
                                  Petsc_False,dmplex_cells,ndim,   &
                                  dmplex_verts,dmda_flow%dmx,ierr)
    CHKERRQ(ierr)
  end if

  if (nprcs > 1) then
    !c Set the flag for creating a mapping to the natural order on distribution
    call DMSetUseNatural(dmda_flow%dmx,PETSC_TRUE,ierr)
    CHKERRQ(ierr)

    !c distribute mesh over processes
    call DMPlexDistribute(dmda_flow%dmx,1,migrationsf,pdm,ierr)
    CHKERRQ(ierr)

    if (pdm /= PETSC_NULL_DM) then
      call DMPlexSetMigrationSF(pdm,migrationsf,ierr)
      CHKERRQ(ierr)

      call DMDestroy(dmda_flow%dmx,ierr)
      CHKERRQ(ierr)

      !c set the global mesh as distributed mesh
      dmda_flow%dmx = pdm
    end if

    !c get number of local cells and nodes
    call DMGetCoordinatesLocal(dmda_flow%dmx,gc,ierr)
    CHKERRQ(ierr)

    call DMGetCoordinateDM(dmda_flow%dmx,cdm,ierr)
    CHKERRQ(ierr)

    call DMGetSection(cdm,section,ierr)
    CHKERRQ(ierr)

    call PetscSectionGetChart(section,istart,iend,ierr)
    CHKERRQ(ierr)

    !c reset local number of nodes and cells after distribution, with ghost nodes/cells
    num_nodes = iend-istart
    num_cells = istart

    write(*,"(3(a,1x,i4,1x))") "rank",rank,"local nodes with ghost",num_nodes,"local cells with ghost",num_cells

    deallocate(nodes)
    deallocate(cells)

    !c set local coordinates
    allocate(nodes(num_nodes))

    !c get coordinates array
    call VecGetArrayF90(gc,coords,ierr)
    CHKERRQ(ierr)

    do ipoint = istart, iend-1

      call PetscSectionGetDof(section,ipoint,dof,ierr)
      CHKERRQ(ierr)

      call PetscSectionGetOffset(section,ipoint,off,ierr)
      CHKERRQ(ierr)

      inode = ipoint-istart+1

      !c set coordinates from 2D (X,Z) back to 3D format (X,Y,Z)
      nodes(inode)%x = coords(off+1)
      nodes(inode)%y = 0.0d0
      nodes(inode)%z = coords(off+2)
    end do

    call VecRestoreArrayF90(gc,coords,ierr)
    CHKERRQ(ierr)

    call PetscSectionDestroy(section,ierr)
    CHKERRQ(ierr)

    !c set local cells list, ignored here

    !c DMDA setup for flow problem, with 1 dof only
    dmda_flow%dim = ndim
    dmda_flow%dof = 1

    numFields = 1
    numComp(1) = 1
    pNumComp => numComp
        
    do i = 1, numFields*(dmda_flow%dim+1)
      numDof(i) = 0
    end do

    !c num of dof for vert, edge, cell
    numDof(0*(dmda_flow%dim+1)+1) = dmda_flow%dof
    pNumDof => numDof
        
    !c no BC is assigned here, use NULL instead for pBcCompIS and pBcPointIS
    numBC = 0
    bcField(1) = 0
    pBcField => bcField

    !c bcCompIS and bcPointIS is not required, use null data
    pBcCompIS => bcCompIS
    pBcPointIS => bcPointIS

    call DMPlexCreateSection(dmda_flow%dmx,dmda_flow%dim,          &
                             numFields,pNumComp,pNumDof,           &
                             numBC,pBcField,                       &
                             pBcCompIS,pBcPointIS,                 &
                             PETSC_NULL_IS,                        &
                             section,ierr)
    CHKERRQ(ierr)
        
    call PetscSectionSetFieldName(section,0,'flow',ierr)
    CHKERRQ(ierr)
        
    call DMSetSection(dmda_flow%dmx,section,ierr)
    CHKERRQ(ierr)

    call PetscSectionDestroy(section,ierr)
    CHKERRQ(ierr)

    call DMSetUp(dmda_flow%dmx,ierr)
    CHKERRQ(ierr)


    !c check global to natural ordering

    !c first generate global vector and assign petsc global vertex order
    call DMCreateGlobalVector(dmda_flow%dmx,vec_global,ierr)
    CHKERRQ(ierr)

    !c Get a pointer to vector data when you need access to the array        
    call VecGetArrayF90(vec_global, vecpointer, ierr)
    CHKERRQ(ierr)

    call DMPlexGetVertexNumbering(dmda_flow%dmx,globalvtx,ierr)
    CHKERRQ(ierr)

    call ISGetIndicesF90(globalvtx,array,ierr)
    CHKERRQ(ierr)

    !c test, set global vector to 1.0
    !scalar = 1.0d0
    !call VecSet(vec_global,scalar,ierr)
    !CHKERRQ(ierr)    

    do inode = 1, num_nodes
      vecpointer(inode) = array(inode)
    end do

    !c Restore the vector when you no longer need access to the array
    call ISRestoreIndicesF90(globalvtx,array,ierr)
    CHKERRQ(ierr)

    !c Restore the vector when you no longer need access to the array
    call VecRestoreArrayF90(vec_global,vecpointer,ierr)
    CHKERRQ(ierr)

    !c second generate natural order and initialize with zero
    call DMCreateGlobalVector(dmda_flow%dmx,vec_natural,ierr)
    CHKERRQ(ierr)

    call VecZeroEntries(vec_natural,ierr)
    CHKERRQ(ierr)

    !c get global to natural order
    call DMPlexGlobalToNaturalBegin(dmda_flow%dmx,vec_global,vec_natural,ierr)
    CHKERRQ(ierr)

    call DMPlexGlobalToNaturalEnd(dmda_flow%dmx,vec_global,vec_natural,ierr)
    CHKERRQ(ierr)

    !c check vec_global
    if (rank == 0) then
      write(*,*) "vector view - global order"
    end if
    call VecView(vec_global,PETSC_VIEWER_STDOUT_WORLD,ierr)
    CHKERRQ(ierr)

    !c check vec_natural
    if (rank == 0) then
      write(*,*) "vector view - natural order"
    end if
    call VecView(vec_natural,PETSC_VIEWER_STDOUT_WORLD,ierr)
    CHKERRQ(ierr)    
  end if

  !c finialize PETSc
  call PetscFinalize(ierr)
  CHKERRQ(ierr)

end program driver_pc

